"""
This file is automatically generated during the setup.
Please do not edit config.py, but edit config.py.in in the root directory.
"""

MPI_INCLUDE_DIRS = "/usr/lib/x86_64-linux-gnu/openmpi/include/openmpi;/usr/lib/x86_64-linux-gnu/openmpi/include/openmpi/opal/mca/event/libevent2022/libevent;/usr/lib/x86_64-linux-gnu/openmpi/include/openmpi/opal/mca/event/libevent2022/libevent/include;/usr/lib/x86_64-linux-gnu/openmpi/include"
MPI_LINK_FLAGS = "-L/usr/lib -pthread"
MPI_LIBRARIES = "/usr/lib/x86_64-linux-gnu/openmpi/lib/libmpi.so"
